'''
a simple templating engine for clorn
'''
'''
VAR_F --> var fragment
OBF --> open block fragment
CBF --> close block fragment
TF --> text fragment
VTS --> var token start
VTE --> var token end
BTS --> block token start
BTE --> block token end
RETOK --> regex token
OLT --> operater lookup table
'''
import re
import operator as oper
import ast
VAR_F=0
OBF=1
CBF=2
TF=3

VTS='{{'
VTE='}}'
BTS='{%'
BTE='%}'

RETOK=re.compile(r"(%s.*?%s|%s.*?%s)"% (VTS, VTE, BTS, BTE))
space=re.compile(r'\s+')
OLT={
	'<':oper.lt,
	'>':oper.gt,
	'==':oper.eq,
	'!=':oper.ne,
	'<=':oper.le,
	'>=':oper.ge
}
class Error(Exception):
	pass
class TemplateNotFoundError(Error):
	pass
class RTempKeyError(Error):
	pass
class RTempSyntaxError(Error):
	pass
def eval(expr):
	try:
		return 'literal', ast.literal_eval(expr)
	except ValueError or SyntaxError:
		return 'name', expr
def resolve(name, context):
	       if name.startswith('..'):
	       	context = context.get('..', {})
	       	name = name[2:]
	       try:
	           for token0 in name.split('.'):
	           	context = context[token0]
	           return context
	       except KeyError as ern:
	       	raise RTempKeyError(f'value for key {ern} is missing') from None
class Fragment(object):
	def __init__(self, text):
		self.text=text
		self.clean=self.clsf()
	def clsf(self):
		if self.text[:2] in (VTS, BTS):
			return self.text.strip()[2:-2].strip()
		return self.text
	@property
	def type(self):
		rs=self.text[:2]
		if rs==VTS:
			return VAR_F
		elif rs==BTS:
			return CBF if self.clean[:3]=='end' else OBF
		else:
			return TF
class NODE(object):
	CS=False
	def __init__(self, f=None):
		self.children=[]
		self.PRFR(f)
	def PRFR(self, fr):
		pass
	def ES(self):
		pass
	def render(self, context):
		pass
	def exit_S(self):
		pass
	def rchild(self, context, ch=None):
		if ch is None:
			ch=self.children
		def rchid(chd):
			chtml=chd.render(context)
			return '' if not chtml else str(chtml)
		return ''.join(map(rchid, ch))
class TNODE(NODE):
	CS=True
class main_(NODE):
	def render(self, context):
		return self.rchild(context)
class VAR(NODE):
	def PRFR(self, frg):
		self.name=frg
	def render(self, context):
		return resolve(self.name, context)
class loop(TNODE):
	def PRFR(self, frg):
		try:
			_, item=space.split(frg, 1)
			self.item=eval(item)
		except ValueError:
			raise SyntaxError(frg)
	def render(self, context):
		items=self.item[1] if self.item[0] == 'literal' else resolve(self.item[1], context)
		def ritem(itm):
			return self.rchild({
				'..':context,
				'item':itm
			})
		return ''.join(map(ritem, items))
class IFCL(TNODE):
	def PRFR(self, fr):
		bits=fr.split()[1:]
		if len(bits) not in (1, 3):
			raise SyntaxError(fr)
		self.lhs=eval(bits[0])
		if len(bits)==3:
			self.op=bits[1]
			self.rhs=eval(bits[2])
	def render(self, context):
		lhs=self.RESIDE(self.lhs, context)
		if hasattr(self, 'op'):
			op=OLT.get(self.op)
			if op ==None:
				raise SyntaxError('')
			rhs=self.RESIDE(self.rhs, context)
			EIB=op(lhs, rhs)
		else:
			EIB=oper.truth(lhs)
		IB, EB=self.SPCH()
		return self.rchild(context, IB if EIB else EB)
	def RESIDE(self, side, context):
		return side[1] if side[0] == 'literal' else resolve(side[1], context)
	def exit_S(self):
		self.IB, self.EB=self.SPCH()
	def SPCH(self):
		IB=[]
		EB=[]
		cur=IB
		for och in self.children:
			if isinstance(och, ELCL):
				cur=EB
				continue
			cur.append(och)
		return IB, EB
class ELCL(NODE):
	def render(self, context):
		pass

class TEXT_(NODE):
	def PRFR(self, fr):
		self.text=fr
	def render(self, context):
		return self.text
class compiler(object):
	def __init__(self, string):
		self.string=string
	def EHFR(self):
		for one in RETOK.split(self.string):
			if one:
				yield Fragment(one)
	def compile(self):
		root=main_()
		SCST=[root]#lister
		for onefr in self.EHFR():
			if not SCST:
				raise TypeError('failed to nest')
			pscope=SCST[-1]
			if onefr.type==CBF:
				pscope.exit_S()
				SCST.pop()
				continue
			newnode=self.CRND(onefr)
			if newnode:
				pscope.children.append(newnode)
				if newnode.CS:
					SCST.append(newnode)
					newnode.ES()
		return root
	def CRND(self, fr):
		nodeclass=None

		if fr.type==TF:
			nodeclass = TEXT_
		elif fr.type==VAR_F:
			nodeclass=VAR
		elif fr.type==OBF:
			try:
				command=fr.clean.split()[0]
			except IndexError:
				raise TypeError('no commands entered') from None
			if command=='loopin':
				nodeclass = loop
			elif command=='if':
				nodeclass = IFCL
			elif command=='else':
				nodeclass=ELCL
		if nodeclass==None:
			raise RTempSyntaxError(f'unknown command "{command}"')
		return nodeclass(fr.clean)
class load(object):
	def __init__(self, string):
		self.string=str(string)
		self.root=compiler(str(string)).compile()
	def render(self, ka):
		return self.root.render(ka)
def rtemp(template, **keys):
	try:
		file=open(template).read()
	except FileNotFoundError:
		file=''
		raise TemplateNotFoundError('no such a template in this directory') from None
	return load(file).render(keys)